;============================================================================
;
; x3dos FAT12 bootloader
;
;============================================================================

;============================================================================
; memory layout
;============================================================================
;
; 0x0000 = Interrupt vector table
; 0x0040 = BIOS data area
; 0x0050 = Kernel load segment
; 0x1000 = Program load segment
; 0x**** = Available memory
; 0x9000 = Root directory               (12288 bytes)
; 0x**** = Available memory
; 0x9600 = 2kb boot stack
; 0x9E00 = Boot sector
; 0xA000 = Video memory
;
;============================================================================

;============================================================================
; kernel entry state
;============================================================================
;
; CS = 0x0050
; DS = 0x0050
; ES = 0x0050
; FS = 0x0050
; GS = 0x0050
; SS = 0x0050
; SP = 0x0000
; IP = 0x0100
;
; CS:IP = program entry
; SS:SP = program stack
;
; AX = loading main address
; BX = loading name pointer
; CX = loading code segment
; DL = boot drive number
;                   
;============================================================================

;============================================================================
; Program entry state
;============================================================================
;
; CS = 0x1000
; DS = 0x1000
; ES = 0x1000
; FS = 0x1000
; GS = 0x1000
; SS = 0x1000
; SP = 0x0000
; IP = 0x0100
;
; CS:IP = program entry
; SS:SP = program stack
;
; AX = loading main address
; BX = loading name pointer
; CX = loading code segment
; DL = boot drive number
;
;============================================================================

;============================================================================
; Bugs start here
;============================================================================

format binary as 'SYS'

use16
org 0

jmp Start                                       ; 0x0000 (2 byte)
nop                                             ; 0x0002 (1 byte)

;============================================================================
; FAT12 data starts here
;============================================================================

BS_OemName              db 'FATBITCH'           ; 0x0003 (8 byte)

;
; BIOS parameter block (BPB)
;

BP_BytesPerSector       dw 0                    ; 0x000B (2 byte)
BP_SectorsPerCluster    db 0                    ; 0x000D (1 byte)
BP_ReservedSectors      dw 0                    ; 0x000E (2 byte)
BP_NumberOfFATs         db 0                    ; 0x0010 (1 byte)
BP_RootEntries          dw 0                    ; 0x0011 (2 byte)
BP_TotalSectors         dw 0                    ; 0x0013 (2 byte)
BP_MediaDescriptor      db 0                    ; 0x0015 (1 byte)
BP_SectorsPerFAT        dw 0                    ; 0x0016 (2 byte)
BP_SectorsPerTrack      dw 0                    ; 0x0018 (2 byte)
BP_HeadsPerCylinder     dw 0                    ; 0x001A (2 byte)
BP_HiddenSectors        dd 0                    ; 0x001C (4 byte)
BP_TotalSectors32       dd 0                    ; 0x0020 (4 byte)

;
; FAT12 data structure
;

BS_DriveNumber          db 0                    ; 0x0024 (1 byte)
BS_Reserved             db 0                    ; 0x0025 (1 byte)
BS_ExtBootSig           db 0                    ; 0x0026 (1 byte)
BS_SerialNumber         dd 0                    ; 0x0027 (4 byte)
BS_VolumeLabel          db 'NO NAME    '        ; 0x002B (11 byte)
BS_FileSystem           db 'FAT12   '           ; 0x0036 (8 byte)

;============================================================================
; Primary boot code starts here
;============================================================================

Start:
        ;====================================================================
        ; clear string direction flag
        ;====================================================================

        cld

        ;====================================================================
        ; how much RAM is there?
        ;====================================================================

        int     0x12                            ; Get conventional memory (in kb)
        shl     ax,6                            ; convert into paragraphs

        ;====================================================================
        ; reserve 512 bytes for boot code
        ;====================================================================

        sub     ax,512/16                       ; subtract 32 paragraphs
        mov     es,ax                           ; ES = top of memory - 512

        ;====================================================================
        ; reserve 2048 bytes for the stack
        ;====================================================================

        sub     ax,2048/16                      ;
        mov     ss,ax                           ; SS = top of memory - 512 - 2048
        mov     sp,2048                         ; Stack is 2048 bytes deep

        ;====================================================================
        ; copy us from DS:SI (0x0000:0x7C00)
        ; to ES:DI (top of memory - 512:0x0000)
        ;====================================================================

        mov     si,0x7C00                       ; SI = 0x7C00
        xor     di,di                           ; DI = 0x0000
        mov     ds,di                           ; DS = 0x0000
        mov     cx,512/2                        ; Number of words to move
        rep     movsw                           ; Copy DS:SI into ES:DI

        ;====================================================================
        ; Setup and jump far into relocated code
        ;====================================================================

        push    es                              ; Top of memory - 512
        push    Start2                          ; Entry point offset
        retf                                    ; Jump to relocated code

;============================================================================
; Relocated boot code starts here
;============================================================================

Start2:

        mov     [cs:BS_DriveNumber],dl

;============================================================================
; Loading code starts here
;============================================================================

Main:

        ;====================================================================
        ; Refresh data segment
        ;====================================================================

        push    cs
        pop     ds

        ;====================================================================
        ; Calculate the FAT size
        ;====================================================================

        mov     ax,[BP_BytesPerSector]          ; AX = bytes per sector
        shr     ax,4                            ; AX = sector size in paragraphs
        mov     cx,[BP_SectorsPerFAT]           ; CX = FAT size in sectors
        mul     cx                              ; AX = FAT size in paragraphs

        ;====================================================================
        ; Setup buffer segment:offset
        ;====================================================================

        mov     di,ss                           ; DI = SS
        sub     di,ax                           ; DI = SS - FAT size in paragraphs
        mov     es,di                           ; ES = SS - FAT size in paragraphs
        xor     bx,bx                           ; ES:BX -> buffer for the FAT

        ;====================================================================
        ; Setup LBA segment:offset
        ;====================================================================

        mov     ax,word[BP_HiddenSectors]       ; AX = hibyte BP_HiddenSectors
        mov     dx,word[BP_HiddenSectors+2]     ; DX = lobyte BP_HiddenSectors
        add     ax,[BP_ReservedSectors]         ; AX = AX + BP_ReservedSectors
        adc     dx,bx                           ; DX:AX = LBA

        ;====================================================================
        ; Load the FAT12 image into buffer
        ;====================================================================
        ; Input: DX:AX = LBA
        ;        CX    = sector count
        ;        ES:BX -> buffer address
        ;====================================================================

        call    ReadSector

        ;====================================================================
        ; Reserve memory for root directory
        ;====================================================================

        mov     bx,ax                           ;
        mov     di,dx                           ; save LBA to DI:BX

        mov     ax,32                           ; AX = 32
        mov     si,[BP_RootEntries]             ; SI = BP_RootEntries
        mul     si                              ; AX = BP_RootEntries * 32
        div     word[BP_BytesPerSector]         ; AX = BP_RootEntries * 32 / BP_BytesPerSector
        mov     cx,ax                           ; CX = root directory size in sectors

        mov     al,[BP_NumberOfFATs]            ; AL = BP_NumberOfFATs
        cbw
        mul     word[BP_SectorsPerFAT]          ;
        add     ax,bx                           ;
        adc     dx,di                           ; DX:AX = LBA

        push    es                              ; push FAT segment (3rd parameter for ReadCluster)

        ;====================================================================
        ; setup ES:BX as BS_ImageLoadSeg:0x0000
        ;====================================================================

        mov     es,[BS_ImageLoadSeg]
        xor     bx,bx                           ; ES:BX -> buffer for root directory

        ;====================================================================
        ; load the root directory
        ;====================================================================
        ; Input: DX:AX = LBA                    
        ;        CX    = sector count           
        ;        ES:BX -> buffer address        
        ;====================================================================

        call    ReadSector

        add     ax,cx                           ;
        adc     dx,bx                           ; adjust LBA for cluster data

        push    dx                              ; setup DX:AX for ReadCluster (2nd parameter for ReadCluster)
        push    ax                              ; push LBA for data (1st parameter for ReadCluster)

    ;========================================================================
    ; Save root directory to safe place (0x9000:0x0000)
    ;========================================================================

        pusha                                   
        push    ds                              
        push    es                              

        ;====================================================================
        ; Setup DS:SI at BS_ImageLoadSeg:0x0000
        ;====================================================================

        mov     ds,[BS_ImageLoadSeg]
        xor     si,si

        ;====================================================================
        ; Setup ES:DI at 0x9000:0x0000
        ;====================================================================

        push    0x9000                          
        pop     es                              
        xor     di,di                           

        ;====================================================================
        ; Copy root directory into memory
        ;====================================================================

        mov     cx,12*1024/2                    
        rep     movsw                           

        pop     es                              
        pop     ds                              
        popa                                    

    ;========================================================================
    ; Setup to find load and run a com file
    ;========================================================================

        mov     di,bx                           ; ES:DI -> root entries array
        mov     dx,si                           ; DX = SI = number of root entries
        mov     si,BS_ProgramName               ; DS:SI -> program name

        ;====================================================================
        ; Look for a file with particular name
        ;====================================================================
        ; Input:  DS:SI -> file name (11 bytes)
        ;         ES:DI -> root directory array
        ;         DX = number of root entries
        ;
        ; Output: SI = cluster number
        ;====================================================================

    FindFileName:
        mov     cx,11                           ; 11 byte file name
    FindNameCycle:                              ;
        cmp     byte[es:di],ch                  ; compare to zero
        je      ErrorNoFile                     ; end of root directory
        pusha                                   ;
        repe    cmpsb                           ;
        popa                                    ;
        je      FindNameFound                   ;
        add     di,32                           ; DI += 32
        dec     dx                              ;
        jnz     FindNameCycle                   ; next root entry
        jmp     ErrorNoFile                     ;
    FindNameFound:                              ;
        mov     si,[es:di+0x1A]                 ; SI = cluster number
    ReadNextCluster:                            ;
        ;
        ; read one cluster (SI = cluster number)
        ;
        mov     bp,sp                           ;
        lea     ax,[si-2]                       ;
        xor     ch,ch                           ;
        mov     cl,[BP_SectorsPerCluster]       ; CX = sector count
        mul     cx                              ;

       ;add     ax,[ss:bp+2]                    ;
       ;adc     dx,[ss:bp+4]                    ; DX:AX = LBA
        add     ax,[ss:bp]
        adc     dx,[ss:bp+2]

        call    ReadSector                      ;
        mov     ax,[BP_BytesPerSector]          ;
        shr     ax,4                            ; AX = paragraphs per sector
        mul     cx                              ; AX = paragraphs read
        mov     cx,es                           ;
        add     cx,ax                           ;
        mov     es,cx                           ; ES:BX updated
        mov     ax,3                            ;
        mul     si                              ;
        shr     ax,1                            ;
        xchg    ax,si                           ; SI = cluster * 3 / 2

        push    ds                              ;
       ;mov     ds,[ss:bp+6]                    ; DS = FAT segment
        mov     ds,[ss:bp+4]
        mov     si,[ds:si]                      ; SI = next cluster
        pop     ds                              ;

        jnc     ReadClusterEven                 ;
        shr     si,4                            ;
    ReadClusterEven:                            ;
        and     si,0x0FFF                       ; mask cluster value

        cmp     si,0x0FF8                       ; is EOF?
        jc      ReadNextCluster                 ; if not EOF then read next cluster

    ;========================================================================
    ; Setup and pass control to a COM program
    ;========================================================================

        cli                                     ; disable interrupt flag (IF = 0)

        mov     ax,[BS_ImageLoadSeg]            ; The kernel loads here once...
        mov     [BS_ImageLoadSeg],0x1010        ; Secondary programs load here...
        sub     ax,0x0010                       ; AX = BS_ImageLoadSeg - 0x0010 = BS_ComLoadSeg

        ;====================================================================
        ; Setup segments and stack
        ;====================================================================

        mov     ds,ax                           ; DS = BS_ComLoadSeg
        mov     es,ax                           ; ES = BS_ComLoadSeg
        mov     fs,ax                           ; FS = BS_ComLoadSeg
        mov     gs,ax                           ; GS = BS_ComLoadSeg
        mov     ss,ax                           ; SS = BS_ComLoadSeg
        xor     sp,sp                           ; clear stack pointer

        ;====================================================================
        ; Pass some parameters to program
        ;====================================================================

        mov     ax,Main                         ; Pass offset of loading code
        mov     bx,BS_ProgramName               ; Pass offset to loading name
        mov     cx,cs                           ; Pass segment of loading code
        mov     dl,[cs:BS_DriveNumber]          ; Pass boot drive number

        ;====================================================================
        ; Jump far into program code
        ;====================================================================

        push    es                              ; push BS_ComLoadSeg
        push    0x0100                          ; push entry point origin
        sti                                     ; enable interrupt flag (IF = 1)
        retf                                    ; Jump far into relocated code

;============================================================================
; Error handler
;
; Uses int 20h, if this is first time boot then it does nothing but hang.
; If we are a secondary program then the kernel will catch 20h and continue.
;============================================================================

ErrorNoFile:
        mov     al,'F'
ErrorNoRead:
        mov     ah,'R'
        int     0x20
        jmp     $

;============================================================================
; Reads one sector into a buffer
;============================================================================
; Input:
;   DX:AX =  LBA
;   CX    =  sector count
;   ES:BX -> buffer address
;
; Output:
;   CF = 1 if error
;   CF = 0 if success
;============================================================================

ReadSector:
        pusha
    ReadSectorNext:
        mov     di,8                            ; number of attempts to read this sector
    ReadSectorRetry:
        pusha
        div     word[BP_SectorsPerTrack]        ; ax = LBA / SPT, dx = LBA % SPT = sector - 1
        mov     cx,dx
        inc     cx                              ; cx = sector number
        xor     dx,dx
        div     word[BP_HeadsPerCylinder]       ; ax = (LBA / SPT) / HPC = cylinder, dx = (LBA / SPT) % HPC = head
        mov     ch,al                           ; ch = LSB 0...7 of cylinder no.
        shl     ah,6
        or      cl,ah                           ; cl = MSB 8...9 of cylinder no. + sector no.
        mov     dh,dl                           ; dh = head number
        mov     dl,[BS_DriveNumber]             ; dl = drive number
        mov     ax,0x0201                       ; HIBYTE(read function = 2) LOBYTE(sector count = 1)
        int     0x13                            ; read sectors
        jc      ResetDisk

        popa
        add     bx,[BP_BytesPerSector]          ; adjust offset for next sector
        add     ax,1
        adc     dx,0                            ; adjust LBA for next sector
        dec     cx
        jnz     ReadSectorNext
        popa
        ret

    ResetDisk:
        xor     ah,ah                           ; ah = 0 = reset function
        int     0x13                            ; reset disk drive
        popa
        dec     di
        jnz     ReadSectorRetry                 ; extra attempt
        jmp     ErrorNoRead



if 0		; few bytes bigger :(
ReadSector:
        pusha                                   
    ReadSectorNext:                             
        mov     di,8                            ; number of attempts to read this sector
    ReadSectorRetry:                            
        pusha                                   
        div     word[BP_SectorsPerTrack]        ; ax = LBA / SPT, dx = LBA % SPT = sector - 1
        mov     cx,dx                           
        inc     cx                              ; cx = sector number
        xor     dx,dx                           
        div     word[BP_HeadsPerCylinder]       ; ax = (LBA / SPT) / HPC = cylinder, dx = (LBA / SPT) % HPC = head
        mov     ch,al                           ; ch = LSB 0...7 of cylinder no.
        shl     ah,6                            
        or      cl,ah                           ; cl = MSB 8...9 of cylinder no. + sector no.
        mov     dh,dl                           ; dh = head number
        mov     dl,[BS_DriveNumber]             ; dl = drive number
        mov     ax,0x0201                       ; HIBYTE(read function = 2) LOBYTE(sector count = 1)
        int     0x13                            ; read sectors
        jnc     ReadSectorDone                  ; CF = 0 if no error
        xor     ah,ah                           ; ah = 0 = reset function
        int     0x13                            ; reset drive
        popa                                    
        dec     di                              
        jnz     ReadSectorRetry                 ; extra attempt
        jmp     ErrorNoRead                     
    ReadSectorDone:                             
        popa                                    
        dec     cx                              
        jz      ReadSectorDone2                 ; last sector
        add     bx,[BP_BytesPerSector]          ; adjust offset for next sector
        add     ax,1                            
        adc     dx,0                            ; adjust LBA for next sector
        jmp     ReadSectorNext                  
    ReadSectorDone2:                            
        popa                                    
        ret                                     
end if

;============================================================================
; General data area
;============================================================================

    ;========================================================================
    ; Loader variables
    ;========================================================================

    BS_ImageLoadSeg  dw 0x0060
    BS_ProgramName   db 'SYSTEM  SYS'

    ;========================================================================
    ; Pad the code and add the boot signature  (size=448)
    ;========================================================================

    times (510-($-$$)) db 0
                       db 0x55
                       db 0xAA

;============================================================================
; Yay!
;============================================================================
