;============================================================================
;
; SYSAPI.ASM
;
; System API code for kernel and user programs
;
;============================================================================



;============================================================================
;
; k_SetVideoMode
;
; Sets the VGA into new video mode
;
; Input: AX = desired video mode
;
; Modes: VIDEOMODE_TEXT
;        VIDEOMODE_GRAPHIC
;
; Todo: Validate mode parameter for stupid coders!
;
;============================================================================

k_SetVideoMode:

        ; AH = set mode function, AL = desired mode

        mov     [cs:g_VideoMode],ax     ; Store it globally

        int     0x10 ; Invoke BIOS service

        retf

;============================================================================
;
; k_GetVideoMode
;
; Gets current video mode
;
; Output: AX = video mode
;
; Modes: VIDEOMODE_TEXT
;        VIDEOMODE_GRAPHIC
;
;============================================================================

k_GetVideoMode:

        mov     ax,[cs:g_VideoMode]

        retf

;============================================================================
;
; k_SetTextColor
;
; Input: BL = desired foreground color
;
; Warning: For text mode only!
;
;============================================================================

k_SetTextColor:

        mov     [cs:g_TextColor],bl

        retf

;============================================================================
;
; k_GetTextColor
;
; Output: BL = current foreground color
;
; Warning: For text mode only!
;
;============================================================================

k_GetTextColor:

        mov     bl,[cs:g_TextColor]

        retf

;============================================================================
;
; k_SetBkgndColor
;
; Input: BL = desired background color
;
; Warning: For text mode only!
;
;============================================================================

k_SetBkgndColor:

        mov     [cs:g_BkgndColor],bl

        retf

;============================================================================
;
; k_GetBkgndColor
;
; Output: BL = current background color
;
; Warning: For text mode only!
;
;============================================================================

k_GetBkgndColor:

        mov     bl,[cs:g_BkgndColor]

        retf

;============================================================================
;
; k_SetWorkArea
;
; Sets work area for clearing and scrolling
;
; Input: CL = left
;        CH = top
;        DL = right
;        DH = bottom
;
; Warning: For text mode only!
;
;============================================================================

k_SetWorkArea:

        ; TODO: Use word instructions!

        mov     [cs:g_WorkArea.Left],cl
        mov     [cs:g_WorkArea.Top],ch
        mov     [cs:g_WorkArea.Right],dl
        mov     [cs:g_WorkArea.Bottom],dh

        retf

;============================================================================
;
; k_GetWorkArea
;
; Gets work area for clearing and scrolling
;
; Output: CL = left
;         CH = top
;         DL = right
;         DH = bottom
;
; Warning: For text mode only!
;
;============================================================================

k_GetWorkArea:

        ; TODO: Use word instructions!

        mov     cl,[cs:g_WorkArea.Left]
        mov     ch,[cs:g_WorkArea.Top]
        mov     dl,[cs:g_WorkArea.Right]
        mov     dh,[cs:g_WorkArea.Bottom]

        retf

;============================================================================
;
; k_ScrollWorkArea
;
; Input: AL = Number of lines to scroll
;
; Scrolls current work area
; Does not update caret position
;
; Warning: For text mode only!
;
;============================================================================

k_ScrollWorkArea:

        pusha

        kcall   k_GetWorkArea   ; Get area into CL,CH,DL,DH
        mov     ah,0x06         ; Scroll window up
        mov     bh,0x0F         ; Color attribute
        int     0x10            ; Invoke BIOS service

        popa

        retf

;============================================================================
;
; k_ClearWorkArea
;
; Clears entire work area
; Sets caret position to top left of work area
;
; Warning: For text mode only!
;
;============================================================================

k_ClearWorkArea:

        pusha

        xor     al,al                   ; Number of lines to scroll (0=all)
        kcall   k_ScrollWorkArea

        mov     dl,[cs:g_WorkArea.Left]    ; x caret position
        mov     dh,[cs:g_WorkArea.Top]     ; y caret position
        kcall   k_SetCaretPos

        popa

        retf

;============================================================================
;
; k_SetCaretPos
;
; Sets caret position
;
; Input:
;   DL = x coordinate (column)
;   DH = y coordinate (row)
;
; Warning: For text mode only!
;
;============================================================================

k_SetCaretPos:

        pusha

        mov     ah,0x02         ; Set caret position
        mov     bh,0x00         ; Page number (0)
        int     0x10            ; Invoke BIOS service

        popa

        retf

;============================================================================
;
; k_GetCaretPos
;
; Gets current caret position
;
; Output: DL = x coordinate (column)
;         DH = y coordinate (row)
;
; Warning: For text mode only!
;
;============================================================================

k_GetCaretPos:

        push    ax
        push    bx
        push    cx              ; Preserve CX (silently modified)

        mov     ah,0x03         ; Sub-function (get caret position)
        mov     bh,0x00         ; Page number index
        int     0x10            ; Invoke BIOS service

        pop     cx
        pop     bx
        pop     ax

        retf

;============================================================================
;
; k_IncCaretPos
;
; Advances caret position
;
; Warning: For text mode only!
;
;============================================================================

k_IncCaretPos:

        push    dx

        kcall   k_GetCaretPos   ; Get current position
        inc     dl              ; Increment column number
        kcall   k_SetCaretPos   ; Apply new caret position

        pop     dx

        retf

;============================================================================
;
; k_DecCaretPos
;
; Retreats caret position
;
; Warning: For text mode only!
;
;============================================================================

k_DecCaretPos:

        push    dx

        kcall   k_GetCaretPos   ; Get current caret position
        dec     dl              ; Decrement column number
        kcall   k_SetCaretPos   ; Apply new caret position

        pop     dx

        retf

;============================================================================
;
; k_PutChar
;
; Prints character at current caret position.
; Does not update caret position.
;
; Input: AL = ASCII character to print
;
; Warning: For text mode only!
;
;============================================================================

k_PutChar:

        pusha

        mov     ah,0x09                 ; Print character function
        mov     bl,[cs:g_BkgndColor]
        or      bl,[cs:g_TextColor]
        xor     bh,bh                   ; Page number (0)
        mov     cx,1                    ; Repeat count
        int     0x10                    ; Invoke BIOS service

        popa

        retf

;============================================================================
;
; k_PrintChar
;
; Input:
;   AL = ASCII character to print
;
; Output:
;   prints character to screen at caret and advances caret position
;
; Warning: For text mode only!
;
;============================================================================

k_PrintChar:

        kcall   k_PutChar       ; Print character
        kcall   k_IncCaretPos   ; Advance caret position

        retf

;============================================================================
;
; k_PrintBuffer
;
; Prints buffer to screen at caret position
;
; Input:
;   DS:SI -> buffer
;   CX = number of bytes to print
;
; Warning: For text mode only!
;
;============================================================================

k_PrintBuffer:

        pusha

    .CharLoop:
        lodsb
        kcall   k_PrintChar
        loop    .CharLoop

        popa

        retf

;============================================================================
;
; k_PrintString
;
; Prints formatted string to screen
;
; Input:                                                                    
;   DS:SI -> string (must be NULL terminated)
;   BL = color attribute (overridden by format)                             
;                                                                           
; Output:                                                                   
;   1) prints string to screen                                              
;   2) caret is automatically advanced                                      
;   3) window is scrolled up if needed 
;
; Warning: For text mode only!
;
; Format: \y - set color to yellow                      
;         \r - set color to red                         
;         \g - set color to green                       
;         \b - set color to blue                        
;         \w - set color to white                       
;         \n - advance caret newline                    
;         \\ - backslash character     
;                 
;============================================================================

k_PrintString:                                  

        pusha

    .reset:
        xor     cx,cx                  		; deactivate sequencer
    .loadc:
        lodsb                                   
        test    al,al                           
        jz      .done                           
        test    cx,cx                           
        jnz     .check                          
        cmp     al,'\'                          
        jne     .putc                           
        inc     cx                    		; CX = 1 (activate sequencer)
        jmp     .loadc                          
    .check:                                     
        ; add new ones here!                    
    .ck_yellow:                                 
        cmp     al,'y'                          
        jne     .ck_red                         
        mov     [cs:g_TextColor],0x0E           ; yellow
        jmp     .reset                          
    .ck_red:                                    
        cmp     al,'r'                          
        jne     .ck_green                       
        mov     [cs:g_TextColor],0x04           ; red
        jmp     .reset                          
    .ck_green:                                  
        cmp     al,'g'                          
        jne     .ck_blue                        
        mov     [cs:g_TextColor],0x02		; green
        jmp     .reset                          
    .ck_blue:                                   
        cmp     al,'b'                          
        jne     .ck_white                       
        mov     [cs:g_TextColor],0x01    	; blue
        jmp     .reset                          
    .ck_white:                                  
        cmp     al,'w'                          
        jne     .ck_endl                        
        mov     [cs:g_TextColor],0x0F           ; white
        jmp     .reset                          
    .ck_endl:                                   
        cmp     al,'n'                          
        jne     .disable                        
        kcall   k_PrintCrlf			; CRLF
        jmp     .reset
    .disable:                                   
        xor     cx,cx                           ; CX = 0 (deactivate sequencer)
    .putc:                                      
        kcall   k_PrintChar
        jmp     .loadc
    .done:
                                    
        popa                                    

	retf

;============================================================================
;
; k_PrintCrlf
;
; Advances caret to next line and scrolls window up if needed
;
; Warning: For text mode only!
;
;============================================================================

k_PrintCrlf:

        pusha

        kcall   k_GetCaretPos
        mov     dl,0                    ; Reset x coordinate
        cmp     dh,24                   ; Has y coordinate hit bottom?
        je      .scroll_up
        inc     dh                      ; Increase y coordinate
        jmp     .set_caret
    .scroll_up:
        mov     al,1                    ; Number of lines to scroll
        kcall   k_ScrollWorkArea        ; Scroll up one line
    .set_caret:
        kcall   k_SetCaretPos

        popa

        retf

;============================================================================
;
; k_PrintPbcd
;
; Unpack and print 2 ASCII characters to screen
;
; Input:
;   AL = packed BCD (PBCD)
;
; Warning: For text mode only!
;
;============================================================================

k_PrintPbcd:

        push    ax

        kcall   k_PbcdToAscii
        kcall   k_PrintChar
        mov     al,ah
        kcall   k_PrintChar

        pop     ax

        retf

;============================================================================
;
; k_UnpackPbcd
;
; Unpacks PBCD into BCD
;
; Input:
;   AL = packed BCD (PBCD)
;
; Output:
;   AX = unpacked BCD
;
;============================================================================

k_UnpackPbcd:

        mov     ah,al           ; Copy into high
        and     ax,0xF00F       ; And it with mask
        shr     ah,4            ; Unpack it

        retf

;============================================================================
;
; k_PbcdToAscii
;
; Convert PBCD into ASCII (2 chars)
;
; Input:
;   AX = packed BCD value (PBCD)
;
; Output:
;   AX = 2 ASCII chars
;
;============================================================================

k_PbcdToAscii:

        kcall   k_UnpackPbcd
        or      ax,0x3030       ; or mask = ASCII
        xchg    al,ah

        retf

;============================================================================
;
; k_CheckForKey
;
; Output:
;   ZF = 1 if key is not available
;   ZF = 0 if key is available
;
;   AH = BIOS scan code
;   AL = ASCII character
;
; Notes: 1) Key is not removed from key buffer
;
;============================================================================

k_CheckForKey:

        mov     ah,0x01         ; Check keystroke
        int     0x16            ; Invoke BIOS service

        retf

;============================================================================
;
; k_WaitForKey
;
; Output: AH = BIOS scan code
;         AL = ASCII character
;
; Notes: 1) Key is removed from key buffer
;
;============================================================================

k_WaitForKey:

        xor     ah,ah           ; Sub-function (get keystroke)
        int     0x16            ; Invoke BIOS service

        retf

;============================================================================
;
; k_SystemPause
;
; Prints paused message and waits for keystroke
;
; Warning: For text mode only!
;
;============================================================================

k_SystemPause:

        pusha
        push    ds

        mov     ax,cs
        mov     ds,ax
        mov     si,g_Paused
        kcall   k_PrintString
        kcall   k_WaitForKey

        pop     ds
        popa

        retf

;============================================================================
;
; k_ExitProgram
;
; Warning: For user program only!
;
;============================================================================

k_ExitProgram:

        xor     ax,ax           ; Internal exit code (see Int20 handler)
        int     0x20
        jmp     $
