;============================================================================
;
; x3dos kernel
;
; Entry:
;
;       AX = Offset of loader main
;       BX = Offset of loader name
;       CX = Segment of loader code
;       DL = Boot drive number
;
;       CS = 0x0050
;       DS = 0x0050
;       ES = 0x0050
;       FS = 0x0050
;       GS = 0x0050
;       SS = 0x0050
;       SP = 0x0000
;       IP = 0x0100
;
;       FAT12 root directory at 0x9000:0x0000
;
;============================================================================

define SYSTEM_BUILD 1
include 'STDAFX.INC'            ; Pull in common preprocs

;============================================================================
; Macro for kernel API farprocs
;============================================================================

macro kcall kproc { call 0x0050:kproc }

;============================================================================
; Macro for adding command handlers
;============================================================================

macro MESSAGE_MAP command*,handler* {
  local marker
  db marker - $ - 1     ; prefixed length
  db command            ; command string
marker:
  dw handler            ; command handler
}

;============================================================================
; Jump past API table into awaiting code
;============================================================================

jmp Start                       ; 0x0050:0x0100

;============================================================================
; System API table
;============================================================================

ApiTableStart:
jmp k_SetVideoMode              ; 0x0050:0x0103
jmp k_SetTextColor              ; 0x0050:0x0106
jmp k_GetTextColor              ; 0x0050:0x0109
jmp k_SetBkgndColor             ; 0x0050:0x010C
jmp k_GetBkgndColor             ; 0x0050:0x010F
jmp k_SetWorkArea               ; 0x0050:0x0112
jmp k_GetWorkArea               ; 0x0050:0x0115
jmp k_ScrollWorkArea            ; 0x0050:0x0118
jmp k_ClearWorkArea             ; 0x0050:0x011B
jmp k_SetCaretPos               ; 0x0050:0x011E
jmp k_GetCaretPos               ; 0x0050:0x0121
jmp k_IncCaretPos               ; 0x0050:0x0124
jmp k_DecCaretPos               ; 0x0050:0x0127
jmp k_PutChar                   ; 0x0050:0x012A
jmp k_PrintChar                 ; 0x0050:0x012D
jmp k_PrintBuffer               ; 0x0050:0x0130
jmp k_PrintCrlf                 ; 0x0050:0x0133
jmp k_PrintPbcd                 ; 0x0050:0x0136
jmp k_UnpackPbcd                ; 0x0050:0x0139
jmp k_PbcdToAscii               ; 0x0050:0x013C
jmp k_CheckForKey               ; 0x0050:0x013F
jmp k_WaitForKey                ; 0x0050:0x0142
jmp k_SystemPause               ; 0x0050:0x0145
jmp k_ExitProgram               ; 0x0050:0x0148
jmp k_PrintString               ; 0x0050:0x014B
ApiTableEnd:

;============================================================================
; Force 3 byte 'jmp Start'
;============================================================================

if ApiTableStart-$$ <> 3
  display 'jmp short Start not allowed!',13,10
end if

repeat 128-($-$$)+2
nop
end repeat

;============================================================================
;
; Kernel entry point (passed from booter)
;
;============================================================================

Start:

    ;========================================================================
    ; Disable interrupts
    ;========================================================================
        cli
    ;========================================================================
    ; Copy parameters passed from booter
    ;========================================================================
        mov     [g_LoaderMain],ax
        mov     [g_LoaderName],bx
        mov     [g_LoaderCode],cx
        add     dl,'A'
        mov     [g_DriveLetter],dl
    ;========================================================================
    ; Setup int 0x20 handler
    ;========================================================================
        push    ds
        xor     ax,ax
        mov     ds,ax
        mov     word[ds:0x20*4],Int20
        mov     word[ds:0x20*4+2],cs
        pop     ds
    ;========================================================================
    ; Enable interrupts
    ;========================================================================
        sti

;============================================================================
;
; Kernel re-entry point (passed from Int20)
;
;============================================================================

Start2:

    ;========================================================================
    ; Set normal string direction
    ;========================================================================
        cld
    ;========================================================================
    ; Enter 80x25 text video mode with fresh screen
    ;========================================================================
        mov      ax,VIDEOMODE_TEXT
        kcall    k_SetVideoMode
    ;========================================================================
    ; Set default text and background colors
    ; (A user program may have changed them)
    ;========================================================================
        mov     [g_TextColor],TEXTCOLOR_WHITE
        mov     [g_BkgndColor],BKGNDCOLOR_BLACK
    ;========================================================================
    ; Print console frame
    ;========================================================================
        mov     si,g_Frame
        kcall   k_PrintString
    ;========================================================================
    ; Insert title into frame
    ;========================================================================
        mov     dx,0x0202               ; y=2,x=2
        kcall   k_SetCaretPos
        mov     si,g_Title              ; Offset of title string
        kcall   k_PrintString
        mov     dx,0x0400               ; y=4,x=0
        kcall   k_SetCaretPos
    ;========================================================================
    ; Setup work area for clearing and scrolling
    ;========================================================================
        mov     [g_WorkArea.Left],0
        mov     [g_WorkArea.Top],4      ; Just below frame
        mov     [g_WorkArea.Right],79
        mov     [g_WorkArea.Bottom],24

;============================================================================
;
; Main loop runs here
;
;============================================================================

MainLoop:

    ;========================================================================
    ; Print command prompt string
    ;========================================================================
        mov     si,g_Prompt                     ; Offset of prompt string
        kcall   k_PrintString
    ;========================================================================
    ; Get command into buffer
    ;========================================================================
        mov     di,g_CmdBuf                     ; Offset of command input buffer
    .GetKey:
        kcall   k_WaitForKey
        cmp     al,SCANCODE_BACKSPACE
        je      .VkBack
        cmp     al,SCANCODE_RETURN
        je      .VkReturn
        cmp     di,c_MaxBuf                     ; Is buffer full?
        je      .GetKey
        stosb                                   ; Append character to buffer
        mov     [g_TextColor],TEXTCOLOR_BLUE
        kcall   k_PrintChar                     ; Print character + advance caret
        jmp     .GetKey
    .VkBack:
        cmp     di,g_CmdBuf                     ; Is buffer empty?
        je      .GetKey
        dec     di                              ; Decrement buffer pointer
        kcall   k_DecCaretPos                   ; Decrement caret position
        xor     al,al                           ; AL = NULL character
        mov     [g_TextColor],TEXTCOLOR_WHITE
        kcall   k_PutChar                       ; Print character (no advance caret)
        jmp     .GetKey
    .VkReturn:
        cmp     di,g_CmdBuf                     ; Is buffer empty?
        je      .GetKey
        mov     byte[di],0                      ; Append NULL terminator to buffer
    ;========================================================================
    ; convert buffer to UPPERCASE (in reverse)
    ;========================================================================
    .ToUpper:                                   ;
        dec     di                              ; Decrement buffer pointer
        cmp     di,g_CmdBuf                     ; Compare to start of buffer
        jb      .CheckCmd
        cmp     byte[di],'a'                    ; < 'a'
        jb      .ToUpper
        cmp     byte[di],'z'                    ; > 'z'
        ja      .ToUpper
        sub     byte[di],32                     ; 32 = 'a' - 'A'
        jmp     .ToUpper
    ;========================================================================
    ; Try find handler for this command
    ;========================================================================
    .CheckCmd:
        mov     si,g_Commands - 2
        xor     cx,cx
    .Compare:
        add     si,cx
        add     si,2
        lodsb
        test    al,al
        jz      .NotCmd
        mov     di,g_CmdBuf
        mov     cl,al
        repe    cmpsb
        jnz     .Compare
        cmp     byte[di],0
        jne     .Compare
        jmp     word[si]
    ;========================================================================
    ; Not a command but maybe a filename?
    ;========================================================================
    .NotCmd:
        mov     si,g_CmdBuf                     ; SI -> command buffer
        mov     cx,9                            ; Search first 9 characters for dot
    .FindDot:
        cmp     byte[si],'.'                    ; Is dot?
        je      .CheckExt
        inc     si                              ; Increment command buffer pointer
        loop    .FindDot
    .BadCmd:
        mov     si,g_BadCmd
        kcall   k_PrintString
        jmp     MainLoop
    .CheckExt:
        ; SI -> '.'
        cmp     word[si+1],0x3358               ; '3' 'X'
        jne     .BadCmd
        cmp     word[si+3],0x0044               ; NULL 'D'
        jne     .BadCmd
    ;========================================================================
    ; Root convert filename
    ; SI should still be pointing at '.'
    ; CX has remainder of filename count
    ;========================================================================
    .PaddOut:                                   ; Padd out remainder...
        mov     byte[si],' '                    ; ...with empty spaces
        inc     si                              ; Increment buffer pointer
        loop    .PaddOut
        mov     word[si-1],0x3358               ; '3' 'X'
        mov     word[si+1],0x0044               ; NULL 'D'
        sub     si,9                            ; point back at filename (g_CmdBuf)
    ;========================================================================
    ; Setup ES,SS,SP for loader
    ;========================================================================
        cli
        int     0x12                            ; get conventional memory size (in KBs)
        shl     ax,6                            ; convert AX into paragraphs
        sub     ax,512/16                       ; reserve 512 bytes for the boot sector code
        mov     es,ax                           ; ES:0x0000 -> top - 512
        sub     ax,2048/16                      ; reserve 2048 bytes for the stack
        mov     ss,ax                           ; SS:0x0000 -> top - 512 - 2048
        mov     sp,2048                         ; 2048 bytes for the stack
        sti
    ;========================================================================
    ; Copy program name into loader name
    ; SI should -> root converted filename
    ;========================================================================
        mov     cx,11                           ; 11 byte filename
        mov     di,[g_LoaderName]               ; Offset of loader name
        rep     movsb                           ; Overwrite loader name
    ;========================================================================
    ; Jump far into code at top of memory
    ;========================================================================
        push    es
        push    [g_LoaderMain]
        retf

;============================================================================
;
; Command handlers
;
;============================================================================

    ;========================================================================
    ; HELP
    ;========================================================================

    OnHelp:

        kcall   k_ClearWorkArea
        mov     si,g_Help               ; Offset of help string
        kcall   k_PrintString
        kcall   k_SystemPause
        kcall   k_ClearWorkArea
        jmp     MainLoop

    ;========================================================================
    ; VER
    ;========================================================================

    OnVer:

        mov     si,g_Ver                ; Offset of version string
        kcall   k_PrintString
        jmp     MainLoop

    ;========================================================================
    ; DATE
    ;========================================================================

    OnDate:

        ;====================================================================
        ; Get system date
        ;====================================================================
        mov     ah,0x04                 ; Get system date
        int     0x1A
        ; DH = month
        ; DL = day of month
        ; CH = century
        ; CL = year
        ;====================================================================
        ; Insert DAY into date string
        ;====================================================================
        mov     al,dl
        kcall   k_PbcdToAscii           ; AX = 2 ASCII characters
        mov     word[g_Date.Day],ax
        ;====================================================================
        ; Insert CENTURY into date string
        ;====================================================================
        mov     al,ch
        kcall   k_PbcdToAscii           ; AX = 2 ASCII characters
        mov     word[g_Date.Century],ax
        ;====================================================================
        ; Insert YEAR into date string
        ;====================================================================
        mov     al,cl
        kcall   k_PbcdToAscii           ; AX = 2 ASCII characters
        mov     word[g_Date.Year],ax
        ;====================================================================
        ; Insert MONTH into date string
        ;====================================================================
        mov     al,dh
        kcall   k_UnpackPbcd            ; One based month index
        dec     ax                      ; Zero based month (0=jan,1=feb,...)
        mov     si,g_Months             ; Offset of month names
        add     si,ax                   ; x 1
        add     si,ax                   ; x 2
        add     si,ax                   ; x 3
        mov     di,g_Date.Month
        mov     cx,3                    ; width (jan,feb,...)
        rep     movsb
        ;====================================================================
        ; Print date string
        ;====================================================================
        mov     si,g_Date               ; Offset to date string
        kcall   k_PrintString
        jmp     MainLoop

    ;========================================================================
    ; TIME
    ;========================================================================

    OnTime:

        ;====================================================================
        ; Get system time
        ;====================================================================
        mov     ah,0x02                 ; Get system time
        int     0x1A
        ; CH = hour
        ; CL = minute
        ; DH = second
        ;====================================================================
        ; Insert AM-PM into time string
        ;====================================================================
        mov     dl,'A'                  ; Assume AM
        cmp     ch,12
        jl      .IsAm
        mov     dl,'P'
    .IsAm:
        mov     byte[g_Time.AmPm],dl
        ;====================================================================
        ; Insert HOUR into time string
        ;====================================================================
        mov     al,ch
        kcall   k_PbcdToAscii           ; AX = 2 ASCII characters
        mov     word[g_Time.Hour],ax
        ;====================================================================
        ; Insert MINUTE into time string
        ;====================================================================
        mov     al,cl
        kcall   k_PbcdToAscii           ; AX = 2 ASCII characters
        mov     word[g_Time.Minute],ax
        ;====================================================================
        ; Insert SECOND into time string
        ;====================================================================
        mov     al,dh
        kcall   k_PbcdToAscii           ; AX = 2 ASCII characters
        mov     word[g_Time.Second],ax
        ;====================================================================
        ; Print time string
        ;====================================================================
        mov     si,g_Time               ; Offset of time string
        kcall   k_PrintString
        jmp     MainLoop

    ;========================================================================
    ; DIR
    ;========================================================================

    OnDir:

        kcall   k_ClearWorkArea
        mov     si,g_Dir                        ; Offset of directory title string
        kcall   k_PrintString

        push    ds                              ; Setup DS:SI at 0x9000:0x0000 (root directory)
        push    0x9000
        pop     ds                              ; DS = 0x9000
        xor     si,si                           ; SI = 0x0000
    .CheckType:
        cmp     byte[ds:si],0xE5                ; If entry has been previously erased and is available
        je      .NextName
        test    byte[ds:si+0x0B],1 shl 3        ; If a volume label
        jnz     .NextName
        cmp     byte[ds:si],0                   ; If entry is available and no subsequent entry is in use
        je      .Finished
        mov     cx,8                            ; 8 byte filename
        kcall   k_PrintBuffer                   ; Prints DS:SI
        mov     al,' '
        kcall   k_PrintChar                     ; Prints AL
        test    byte[ds:si+0x0B],1 shl 4        ; If a sub-directory
        jz      .NotSubDir
        mov     al,'D'
        kcall   k_PrintChar                     ; Prints AL
        mov     al,'I'
        kcall   k_PrintChar
        mov     al,'R'
        kcall   k_PrintChar                     ; Prints AL
        kcall   k_PrintCrlf                     ; Prints CRLF
        jmp     .NextName
    .NotSubDir:
        add     si,8                            ; Point at extension
        mov     cx,3                            ; 3 byte extension
        kcall   k_PrintBuffer                   ; Prints DS:SI
        kcall   k_PrintCrlf                     ; Prints CRLF
        sub     si,8                            ; Point back at filename
    .NextName:
        add     si,32                           ; Advance to next filename
        jmp     .CheckType
    .Finished:
        pop     ds                              ; Must be restored ASAP!

        kcall   k_SystemPause
        kcall   k_ClearWorkArea
        jmp     MainLoop

    ;========================================================================
    ; CLS
    ;========================================================================

    OnCls:

        kcall   k_ClearWorkArea
        jmp     MainLoop

    ;========================================================================
    ; REBOOT
    ;========================================================================

    OnReboot:

        mov     si,g_Reboot     ; Offset of reboot message
        kcall   k_PrintString   ; Print rebooting message
        kcall   k_SystemPause   ; Print paused message and await keypress

        db      0xEA            ; Machine language jumps
        dw      0x0000          ; to address FFFF:0000
        dw      0xFFFF          ; and reboots the system
        jmp     $               ; Just in case :S

;============================================================================
;
; Int20
;
; Handles interrupt 0x20 (Terminate X3D program)
;
; Input:
;   AX = exit code
;
; Exit Codes:
;   AL = 'F' If failed to find file
;   AH = 'R' If failed to read file
;   AX =  0  If reporting success
;
;============================================================================

Int20:

        ;====================================================================
        ; Refresh special registers
        ;====================================================================
        cli
        mov     dx,cs
        mov     ds,dx
        mov     es,dx
        mov     fs,dx
        mov     gs,dx
        mov     ss,dx
        xor     sp,sp
        sti
        ;====================================================================
        ; Check exit code and resume kernel
        ;====================================================================
        cmp     al,'F'          ; Failed to find file?
        je      .NoFindFile
        cmp     ah,'R'          ; Failed to read file?
        jne     Start2          ; If no error, re-enter kernel at 'Start2'
    ;NoReadFile
        mov     si,g_NoRead     ; Offset of error string
        jmp     .PrintError
    .NoFindFile:
        mov     si,g_NoFind     ; Offset of error string
    .PrintError:
        kcall   k_PrintString
        jmp     MainLoop        ; If loader failure, re-enter at MainLoop

;============================================================================
; Pull in API code
;============================================================================

; Force 3 byte far jumps

if $-ApiTableEnd <= 128
display 'API table too close!'
end if

;repeat 128-($-ApiTableEnd)+2
;nop
;end repeat

include 'SYSAPI.ASM' ; Pull in API code

;============================================================================
;
; Data starts here
;
;============================================================================

    ;========================================================================
    ; Bootloader parameters
    ;========================================================================

    g_LoaderMain dw 0
    g_LoaderName dw 0
    g_LoaderCode dw 0

    ;========================================================================
    ; Video attributes
    ;========================================================================

    g_VideoMode  dw 0
    g_TextColor  db 0
    g_BkgndColor db 0

    ;========================================================================
    ; Work area for clearing and scrolling
    ;========================================================================

    g_WorkArea:
      .Left   db 0
      .Top    db 0
      .Right  db 0
      .Bottom db 0

    ;========================================================================
    ; Console input buffer
    ;========================================================================

    g_CmdBuf db 8+1+3+1 dup(0) ; filename,dot,ext,NULL
    c_MaxBuf = $ - 1

    ;========================================================================
    ; Console commands and handlers
    ;========================================================================

    g_Commands:
        MESSAGE_MAP 'HELP',   OnHelp
        MESSAGE_MAP 'VER',    OnVer
        MESSAGE_MAP 'DATE',   OnDate
        MESSAGE_MAP 'TIME',   OnTime
        MESSAGE_MAP 'DIR',    OnDir
        MESSAGE_MAP 'CLS',    OnCls
        MESSAGE_MAP 'REBOOT', OnReboot
        db 0 ; SENTINEL

    ;========================================================================
    ; General strings
    ;========================================================================

    g_NoFind            db '\n\rFailed to find file.\n',0
    g_NoRead            db '\n\rFailed to read file.\n',0
    g_Title             db '\wType \yHELP \wif you need it...',0
    g_BadCmd            db '\n\rBad command or file name.\n',0
    g_Paused            db '\wPress any key to continue...',0
    g_Ver               db '\n\gCurrent version is \w0.0.0.1 (alpha)\n',0
    g_Dir               db '\gDirectory contents:\y\n',0
    g_Reboot            db '\n\gRebooting machine.\n',0

    ;========================================================================
    ; Console frame string
    ;========================================================================

    g_Frame:
        db '\b                                                                 Ŀ      \n'
        db 'Ĵ \gx3dos\b Ŀ\n'
        db '                                                                     \n'
        db '\n'
        db 0

    ;========================================================================
    ; Help menu string
    ;========================================================================

    g_Help:
        db '\wThanks for trying x3dos!\n'
        db '\gConsole command list:\n'
        db '\yHELP   \w- Display this screen.\n'
        db '\yVER    \w- Display current version.\n'
        db '\yDATE   \w- Display current date.\n'
        db '\yTIME   \w- Display current time.\n'
        db '\yDIR    \w- Display root directory.\n'
        db '\yCLS    \w- Clears console screen.\n'
        db '\yREBOOT \w- Reboots the machine.\n'
        db '\gMore to come soon...\n'
        db 0

    ;========================================================================
    ; Command prompt string (holds drive letter)
    ;========================================================================

    g_Prompt:
        db '\w'                 ; White text
    g_DriveLetter:
        db 'Z'                  ; Default drive letter
        db ':'
        db '\\'                 ; Backslash (escape an escape sequence)
        db 0

    ;========================================================================
    ; Time string with insertion points
    ;========================================================================

    g_Time:
        db '\n'                 ; New line
        db '\g'                 ; Green text
        db 'Current time is '
        db '\w'                 ; White text
      .Hour:
        db '??'
        db ':'
      .Minute:
        db '??'
        db ':'
      .Second:
        db '??'
        db ' '
      .AmPm:                    ; 'A' or 'P'
        db '?M'
        db '\n'                 ; New line
        db 0

    ;========================================================================
    ; Date string with insertion points
    ;========================================================================

    g_Date:
        db '\n'                 ; New line
        db '\g'                 ; Green text
        db 'Current date is '
        db '\w'                 ; White text
      .Month:
        db '???'
        db '/'
      .Day:
        db '??'
        db '/'
      .Century:
        db '??'
      .Year:
        db '??'
        db '\n'                 ; New line
        db 0

    ;========================================================================
    ; 3 byte g_Date.Month names
    ;========================================================================

    g_Months:
        db 'jan'
        db 'feb'
        db 'mar'
        db 'apr'
        db 'may'
        db 'jun'
        db 'jul'
        db 'aug'
        db 'sep'
        db 'oct'
        db 'nov'
        db 'dec'

;============================================================================
; Yay!
;============================================================================
